#! /usr/bin/wish
#
# hexrev.tcl : ヘックスリバーシ
#
#              Copyright (C) 2005,2014,2019 Makoto Hiroi
#
#	盤面の座標（１次元配列で表す）
#
#		　０　１　２　３　４　５
#		６  ７  ８  ９  10  11  12
#	      13  14  15  16  17  18  19  20
#           21  22  23  24  25  26  27  28  29
#	  30  31  32  33  34  35  36  37  38  39
#	40  41  42  43  44  45  46  47  48  49  50
#	  51  52  53  54  55  56  57  58  59  60
#	    61  62  63  64  65  66  67  68  69
#	      70  71  72  73  74  75  76  77
#		78  79  80  81  82  83  84
#		  85  86  87  88  89  90
#
# 大域変数
# piece()   : 駒を表す図形 id を格納
# action    : 0 : 先手, 1 : 後手
# color     : 駒の色
# play_flag : ゲーム進行用フラグ 1 :先手（黒）, 2 : 後手（白）
# level     : 思考レベル
# message   : メッセージ表示用ラベル
# score     : 玉数表示用ラベル
# last_score: 最終スコア予想
# black_piece : 黒の残り（DLL 側でセット）
# white_piece : 白の残り（DLL 側でセット）
# VALUE       : 評価値（DLL 側でセット）
#
# reversi.dll で追加されるコマンド
#
# init_game                  : ゲームの初期化
# think $level $turn $random : 思考ルーチン
# reverse $n $turn           : ひっくり返る位置をリストで返す
# check_reverse $n $turn     : 石を裏返すことができるか
# count_place $turn          : 石を置ける場所の数
#

# グラフィックを表示する座標
set zahyou {
  162  70 200  70 238  70 276  70 314  70 352  70
  143 103 181 103 219 103 257 103 295 103 333 103 371 103 
  124 136 162 136 200 136 238 136 276 136 314 136 352 136 390 136 
  105 169 143 169 181 169 219 169 257 169 295 169 333 169 371 169 409 169 
   86 202 124 202 162 202 200 202 238 202 276 202 314 202 352 202 390 202 428 202
   67 235 105 235 143 235 181 235 219 235 257 235 295 235 333 235 371 235 409 235 447 235 
   86 268 124 268 162 268 200 268 238 268 276 268 314 268 352 268 390 268 428 268 
  105 301 143 301 181 301 219 301 257 301 295 301 333 301 371 301 409 301 
  124 334 162 334 200 334 238 334 276 334 314 334 352 334 390 334 
  143 367 181 367 219 367 257 367 295 367 333 367 371 367 
  162 400 200 400 238 400 276 400 314 400 352 400 
}

# 思考ルーチンロード
load ./hexrev.so

# ヘルプファイルの表示
proc help {} {
    global argv0
    if {![winfo exist .t0]} {
        toplevel .t0
        wm title .t0 "HexReversi Help"
        text .t0.text -yscrollcommand ".t0.scroll set"
        scrollbar .t0.scroll -command ".t0.text yview"
	pack .t0.scroll -side right -fill y
	pack .t0.text -side left
	# 起動パスを取得
	set path [file dirname $argv0]
	# ファイルの読み込み
	set f [open "$path/hexrev.txt" r]
	while {![eof $f]} {
	    .t0.text insert end [read $f 1000]
	}
	close $f
	.t0.text configure -state disabled
    }
}

# 駒を描画する
proc draw_piece {n c} {
    global piece color
    .c0 itemconfigure $piece($n) -fill $color($c)
    .c0 raise $piece($n)
}

# 駒を置く
proc put_piece {piece_list turn} {
    foreach n $piece_list {
	draw_piece $n $turn
	update
	after 250
    }
    display_score
}

# メッセージの表示
proc display_message {m} {
    global message mes_table VALUE
    set message $mes_table($m)
    update
    after 500
}

# スコアの表示
proc display_score {} {
    global score white_piece black_piece
    set score [format "黒 %2d 石 : 白 %2d 石" $black_piece $white_piece]
    update
}

# 予想スコアの表示
proc display_last_score {} {
    global last_score VALUE
    set result [expr abs($VALUE)]
    if {$VALUE > 0} {
	.ll configure -fg black
	set m [format "黒の %2d 石勝ちでしょう" $result]
    } elseif {$VALUE < 0} {
	.ll configure -fg white
	set m [format "白の %2d 石勝ちでしょう" $result]
    } else {
	set m "引き分けでしょう"
    }
    set last_score $m
    update
}

# 終了ダイアログ    
proc game_over {} {
    global play_flag action white_piece black_piece
    set c [expr $black_piece - $white_piece]
    if {$c == 0} {
	set m "引き分けです"
    } elseif {($action == 0 && $c > 0) || ($action == 1 && $c < 0)} {
	set m "あなたの勝ちです"
    } else {
	set m "私の勝ちです"
    }
    tk_messageBox -type ok -message $m
    set play_flag 0
    display_message start
}

# 探索レベルの決定
proc decide_level {level} {
    global white_piece black_piece
    set rest [expr 91 - $white_piece - $black_piece]
    # レベル５で残り 14 手で読み切り
    if {[expr 9 + $level] >= $rest} {
	# 読み切りモード
	display_message finish
	return 20
    }
    return $level
}

# コンピュータ側の手番
proc put_piece_com {} {
    global play_flag level action white_piece black_piece random
    global VALUE
    while {1} {
	. configure -cursor watch
	display_message com_turn
	set turn [expr $play_flag - 1]
	set lv [decide_level $level] 
	set result [think $lv $turn $random]
	if {$lv == 20} {
	    display_last_score
	}
	. configure -cursor ""
	put_piece $result $turn 
	# 状態チェック
	if {[count_place $action] == 0} {
	    if {[count_place $turn] == 0} {
		game_over
		return
	    } else {
		# 相手がパスだよ
		display_message pass
	    }
	} else {
	    break
	}
    }
    # 手番を移す
    if {$play_flag == 1} {
	set play_flag 2
    } else {
	set play_flag 1
    }
    display_message my_turn
}

# 人間側の指し手
proc put_piece_human {n} {
    global play_flag white_piece black_piece action message
    # 自分の手番かチェックする
    if {$play_flag != [expr $action + 1]} return
    # 駒を反転できるかチェックする
    if ![check_reverse $n $action] return
    # 二重入力防止
    set $play_flag -1
    # 駒の位置がリストで返される
    put_piece [reverse $n $action] $action
    # 状態チェック
    if {[count_place [expr !$action]] == 0} {
	if {[count_place $action] == 0} {
	    # 両方とも打つ場所無し
	    game_over
	    return
	} else {
	    # 相手はパスだよ
	    display_message pass
	    display_message my_turn
	    return
	}
    }
    # 元に戻す
    set play_falg [expr $action + 1]

    # 手番を移す
    if {$play_flag == 1} {
	set play_flag 2
    } else {
	set play_flag 1
    }
    display_message com_turn
    put_piece_com
}

# 中断ダイアログ
proc break_dialog {} {
    set ans [tk_messageBox -type yesno -icon question \
	                   -message "今のゲームを中断しますか？"]
    return [expr {($ans == "no") ? 0 : 1}]
}

# 駒を隠す
proc hide_piece {} {
    global piece
    for {set i 0} {$i < 91} {incr i} {
	.c0 lower $piece($i)
    }
}

# ゲームの開始
proc new_game {} {
    global play_flag action white_piece black_piece last_score
    if {$play_flag != 0} {
	if ![break_dialog] return
	set play_flag 0
    }
    hide_piece
    if {$play_flag == 0 } {
	# ゲーム開始
	init_game
	set white_piece 2
	set black_piece 2
	# 最初の駒を描画
	draw_piece 34 1
	draw_piece 35 0
	draw_piece 44 0
	draw_piece 45 0
	draw_piece 46 1
	draw_piece 55 1
	draw_piece 56 0
	set play_flag 1
	set last_score ""
	display_score
	if $action {
	    put_piece_com
	} else {
	    display_message my_turn
	}
    }
}

# グローバル変数の初期化
proc init_global {} {
    global level action mes_table play_flag message color random
    set level 1
    set action 0
    set random 1
    set play_flag 0
    set color(0) black
    set color(1) white
    set mes_table(start)    "メニュー Games の Start でゲーム開始！"
    set mes_table(my_turn)  "あなたの手番です"
    set mes_table(com_turn) "わたしの手番です"
    set mes_table(pass)     "・・・パスです・・・"
    set mes_table(finish)   "最後まで読み切ります！"
    set message $mes_table(start)
}

# ********** メニューの設定 **********
menu .m -type menubar
. configure -menu .m
.m add cascade -label "Games" -under 0 -menu .m.m1
.m add cascade -label "Level" -under 0 -menu .m.m2
.m add command -label "Help"  -under 0 -command "help"
menu .m.m1 -tearoff no
.m.m1 add command -label "Start" -under 0 -command "new_game"
.m.m1 add separator
.m.m1 add radiobutton -label "先手" -variable action -value 0
.m.m1 add radiobutton -label "後手" -variable action -value 1
.m.m1 add separator
.m.m1 add checkbutton -label "Rand" -variable random
.m.m1 add separator
.m.m1 add command -label "Exit" -under 0 -command "exit"
menu .m.m2 -tearoff no
.m.m2 add radiobutton -label "Level 0" -variable level -value 0
.m.m2 add radiobutton -label "Level 1" -variable level -value 1
.m.m2 add radiobutton -label "Level 2" -variable level -value 2
.m.m2 add radiobutton -label "Level 3" -variable level -value 3
.m.m2 add radiobutton -label "Level 4" -variable level -value 4
.m.m2 add radiobutton -label "Level 5" -variable level -value 5

# 画面の初期化
option add *font "{Noto Sans Mono CJK JP} 12"
canvas .c0 -height 390 -width 430 -background green4

# 六角形を描画する
set dx {  0 -19 -19  0 19  19}
set dy {-22 -11  11 22 11 -11}
for {set i 0} {$i < 91} {incr i} {
    set x [expr [lindex $zahyou [expr $i * 2]] - 40]
    set y [expr [lindex $zahyou [expr $i * 2 + 1]] - 40]
    set l ""
    for {set j 0} {$j < 6} {incr j} {
	set l [concat $l [expr $x + [lindex $dx $j]] [expr $y + [lindex $dy $j]]]
    }
    set k [eval ".c0 create polygon $l -outline black -fill green4" -width 2.0]
    .c0 bind $k <Button-1> "put_piece_human $i"
    # 石の描画
    set x1 [expr $x - 15]
    set y1 [expr $y - 15]
    set x2 [expr $x + 15]
    set y2 [expr $y + 15]
    set piece($i) [.c0 create oval $x1 $y1 $x2 $y2 -outline green4]
    .c0 lower $piece($i)
}

# メッセージ表示用ラベル
label .lm -textvariable message -bg green4 -fg yellow
# スコア表示用
label .ls -textvariable score -bg green4 -fg white
# 予想スコア
label .ll -textvariable last_score -bg green4 -fg black
pack .ls .lm .ll .c0 -fill x

wm title . "Hex Reversi"
wm resizable . 0 0

init_global
focus -force .

# end of file

